unit hiFileStructReader;

interface

uses Kol,Share,Debug,RegExpr;

const
  STYPE_INVALID = 0;
  STYPE_INTEGER = 1;
  STYPE_REAL    = 2;
  STYPE_BYTE    = 3;
  STYPE_WORD    = 4;
  STYPE_CHAR    = 5;
  STYPE_STRING  = 6;
  STYPE_ZSTRING = 7;

type
  TStructInfo = record
    ftype: byte;
    isArray: boolean;
    isSigned: boolean;
    isBigEndian: boolean;
    argSize: byte;
    arrlen: cardinal;
  end;
  
  THIFileStructReader = class(TDebug)
   private
    Struct: array of TStructInfo;
    StructDt: array of TData;
    procedure SetStructure(Value: string);
    procedure ReadField(st:PStream; var dt:TData; ftype:TStructInfo);
   public
    onRead:THI_Event;
    stream:THI_Event;

    destructor Destroy; override;
    procedure doRead(var _Data:TData; Index:word);
    procedure _var_Structure(var _Data:TData; Index:word);
    
    property _prop_Structure:string write SetStructure;
  end;
  
function Str2StructureType(str:string):byte;
procedure ReadStructure(var Struct: array of TStructInfo; list: PStrList);
procedure ReverseBytes(p: Pointer; dtSize: byte);

implementation

function Str2StructureType;
begin
  str := LowerCase(str);
  if str = 'integer' then Result := STYPE_INTEGER
  else if str = 'real' then Result := STYPE_REAL
  else if str = 'byte' then Result := STYPE_BYTE
  else if str = 'word' then Result := STYPE_WORD
  else if str = 'char' then Result := STYPE_CHAR
  else if str = 'string' then Result := STYPE_STRING
  else if str = 'zstring' then Result := STYPE_ZSTRING
  else Result := STYPE_INVALID;
end;

procedure ReverseBytes;
var pb, pe: ^byte;
    b: byte;
begin
   pb := p; pe := pb; inc(pe, dtSize-1); 
   while cardinal(pb) < cardinal(pe) do begin
      b := pb^; pb^ := pe^; pe^ := b;
      inc(pb); dec(pe);
   end;
end;

// ^([\w\s]+)=\s*(\w+?)(_([su])?([lb]e)?)?\s*(:\s*(\d+)\s*)?(\[\s*(\d+)\s*\])?\s*$
// 1: varname
// 2: type
// 4: [s]igned/[u]nsigned
// 5: little/big
// 7: argsize
// 9: arrsize
procedure ReadStructure;
var i: integer;
    re: TRegExpr;
begin
  re := TRegExpr.Create;
  re.ModifierI := True;
  re.ModifierG := True;
  re.Expression := '^([\w\s]+)=\s*(\w+?)(_([su])?([lb]e)?)?\s*(:\s*(\d+)\s*)?(\[\s*(\d+)\s*\])?\s*$';
  for i := 0 to list.Count-1 do begin
    if re.Exec(list.Items[i]) then begin
      Struct[i].ftype := Str2StructureType(re.Match[2]);
      
      Struct[i].isSigned := True;
      Struct[i].argSize := 4;
      
      if (re.SubExprMatchCount >= 4) and (re.MatchPos[4] >= 0) then begin
        Struct[i].isSigned := re.Match[4] = 's';
      end;
      
      if (re.SubExprMatchCount >= 5) and (re.MatchPos[5] >= 0) then begin
        Struct[i].isBigEndian := re.Match[5] = 'be';
      end else begin
        Struct[i].isBigEndian := False;
      end;
      
      if (re.SubExprMatchCount >= 7) and (re.MatchPos[7] >= 0) then begin
        Struct[i].argSize := Str2Int(re.Match[7]);
      end;
      
      case Struct[i].ftype of
        STYPE_INTEGER: begin
          if Struct[i].argSize < 1 then Struct[i].argSize := 1
          else if Struct[i].argSize > 8 then Struct[i].argSize := 8;
        end;
        
        STYPE_BYTE: begin
          Struct[i].isSigned := False;
          Struct[i].argSize := 1;
        end;
        
        STYPE_WORD: begin
          Struct[i].isSigned := False;
          Struct[i].argSize := 2;
        end;
        
        STYPE_CHAR: begin
          Struct[i].argSize := 1;
        end;
        
        STYPE_REAL: begin
          if Struct[i].argSize <= 4 then Struct[i].argSize := 4
          else if Struct[i].argSize <= 8 then Struct[i].argSize := 8
          else Struct[i].argSize := 10;
        end;
        
        STYPE_STRING: begin
          if Struct[i].argSize < 1 then Struct[i].argSize := 1
          else if Struct[i].argSize > 4 then Struct[i].argSize := 4;
        end;
      end;
      
      if re.SubExprMatchCount >= 9 then begin
        Struct[i].isArray := True;
        Struct[i].arrlen := Str2Int(re.Match[9]);
      end else begin
        Struct[i].isArray := False;
      end;
      
//      _debug(re.Match[1] + '=' + re.Match[2] + '_' + re.Match[4] + Int2Str(integer(Struct[i].isBigEndian)) + '(' + Int2Str(Struct[i].ftype) + '):'
//             + Int2Str(Struct[i].argSize) + '[' + Int2Str(Struct[i].arrlen) + '], ' + Int2Str(integer(Struct[i].isArray)));
    end;
    //if re.LastError <> 0 then _debug(RE.ErrorMsg(re.LastError));
  end;
  re.Free;
end; 

destructor THIFileStructReader.Destroy;
var i: integer;
begin
  for i := 0 to high(StructDt) do begin
    if StructDt[i].ldata <> nil then
      FreeData(StructDt[i].ldata);
  end;
  inherited;
end;

procedure THIFileStructReader.SetStructure;
var list: PStrList;
begin
  list := NewStrList;
  list.Text := Value;
  SetLength(Struct, list.Count);
  SetLength(StructDt, list.Count);
  ReadStructure(Struct, list);
  list.Free;
end;

procedure THIFileStructReader.ReadField;
type dtrec = record
       case byte of
         0: (idt: int64);
         1: (cdt: char);
         2: (sdt: single);
         3: (rdt: real);
         4: (edt: extended);
     end;
var fdt: dtrec;
    idt: int64;
    str: string;
    cnt: cardinal;
begin
  //st, dt, ftype
  case ftype.ftype of
    STYPE_BYTE,
    STYPE_WORD,
    STYPE_INTEGER: begin
      fdt.idt := 0;
      st.Read(fdt.idt, ftype.argSize);
      if ftype.isBigEndian then
        ReverseBytes(@fdt.idt, ftype.argSize);
      if ftype.isSigned then begin
         idt := (int64(1) shl (ftype.argSize * 8));
         if ((idt shr 1) and fdt.idt) <> 0 then dec(fdt.idt, idt);
      end; 
      if (fdt.idt <= MAXINT) and (fdt.idt >= -MAXINT-1) then
        dtInteger(dt, fdt.idt)
      else
        dtReal(dt, fdt.idt);
    end;
    
    STYPE_REAL: begin
      st.Read(fdt.edt, ftype.argSize);
      if ftype.isBigEndian then
        ReverseBytes(@fdt.edt, ftype.argSize);
      case ftype.argSize of
         4: dtReal(dt, fdt.sdt);
         8: dtReal(dt, fdt.rdt);
       else dtReal(dt, fdt.edt);
      end;
    end;
    
    STYPE_CHAR: begin
      st.read(fdt.cdt, sizeof(char));
      dtString(dt, fdt.cdt);
    end;
    
    STYPE_STRING: begin
      cnt := 0;
      st.read(cnt, ftype.argSize);
      if ftype.isBigEndian then
        ReverseBytes(@cnt, ftype.argSize);
      SetLength(str, cnt);
      st.read(str[1], cnt);
      dtString(dt, str);
    end;
    
    STYPE_ZSTRING: begin
      dtString(dt, st.ReadStrZ);
    end;
    
    else begin
      dtNull(dt);
    end;
  end
end;

procedure THIFileStructReader.doRead;
var st: PStream;
    i, j: integer;
    dt, tdt: TData;
    str: string;
    mt: PMt;
begin
  dt := _Data;
  st := ReadStream(_Data, stream, nil);
  if st <> nil then begin
    for i := 0 to high(Struct) do begin
      if StructDt[i].ldata <> nil then
        FreeData(StructDt[i].ldata);
      if (Struct[i].isArray) and (Struct[i].arrlen > 1) then begin
        if Struct[i].ftype = STYPE_CHAR then begin
          SetLength(str, Struct[i].arrlen);
          st.read(str[1], Struct[i].arrlen);
          dtString(StructDt[i], str);
        end else begin
          mt := mt_make(StructDt[i]);
          ReadField(st, StructDt[i], Struct[i]);
          for j := 1 to Struct[i].arrlen-1 do begin
            ReadField(st, tdt, Struct[i]);
            mt_data(mt, tdt);
          end;
          dispose(mt);
        end;
      end else begin
        ReadField(st, StructDt[i], Struct[i]);
      end;
    end;
    _hi_CreateEvent(_Data, @onRead, dt);
  end;
end;

procedure THIFileStructReader._var_Structure;
begin
  _Data := StructDt[Index];
end;

end.
