unit hiFileStructWriter;

interface

uses Kol,Share,Debug, hiFileStructReader;

type
  THIFileStructWriter = class(TDebug)
   private
    strmidx: integer;
    Struct: array of TStructInfo;
   
    procedure SetStructure(Value:string);
    procedure WriteField(st:PStream; var dt:TData; ftype:TStructInfo);
   public
    _data_structure:array of THI_Event;
    onWrite:THI_Event;

    constructor Create;
    destructor Destroy; override;
    procedure doWrite(var _Data:TData; Index:word);
    
    property _prop_Structure:string write SetStructure;
  end;

implementation

constructor THIFileStructWriter.Create;
begin
  strmidx := -1;
  inherited;
end;

destructor THIFileStructWriter.Destroy;
begin
  
  inherited;
end;

procedure THIFileStructWriter.WriteField;
type dtrec = record
       case byte of
         0: (idt: int64);
         1: (cdt: char);
         2: (sdt: single);
         3: (rdt: real);
         4: (edt: extended);
     end;
var fdt: dtrec;
    str: string;
    cnt: cardinal;
begin
  //st, dt, ftype
  case ftype.ftype of
    STYPE_WORD,
    STYPE_BYTE,
    STYPE_INTEGER: begin
      if ftype.argSize < 5 then
        fdt.idt := ToInteger(dt)
      else
        fdt.idt := Round(ToReal(dt));
      if ftype.isBigEndian then
        ReverseBytes(@fdt.idt, ftype.argSize);
      st.Write(fdt.idt, ftype.argSize);
    end;
    
    STYPE_REAL: begin
      case ftype.argSize of
         4: fdt.sdt := ToReal(dt);
         8: fdt.rdt := ToReal(dt);
       else fdt.edt := ToReal(dt);
      end;
      if ftype.isBigEndian then
        ReverseBytes(@fdt.edt, ftype.argSize);
      st.Write(fdt.edt, ftype.argSize);
    end;
    
    STYPE_CHAR: begin
      fdt.cdt := ToString(dt)[1];
      st.write(fdt.cdt, sizeof(char));
    end;
    
    STYPE_STRING: begin
      str := ToString(dt);
      cnt := Length(str);
      cnt := cnt and ((1 shl ftype.argSize * 8) - 1);
      fdt.idt := cnt;
      if ftype.isBigEndian then
        ReverseBytes(@cnt, ftype.argSize);
      st.write(cnt, ftype.argSize);
      st.write(str[1], fdt.idt);
    end;
    
    STYPE_ZSTRING: begin
      st.WriteStrZ(ToString(dt));
    end;
    
    else begin
      
    end;
  end
end;

procedure THIFileStructWriter.SetStructure;
var list: PStrList;
begin
  list := NewStrList;
  list.Text := Value;
  SetLength(_data_structure, list.Count);
  strmidx := list.IndexOf('#stream');
  if strmidx >= 0 then begin
    list.Delete(strmidx);
  end;
  SetLength(Struct, list.Count);
  ReadStructure(Struct, list);
  list.Free;
end;

procedure THIFileStructWriter.doWrite;
var st: PStream;
    i, si, j: integer;
    dt, tdt: TData;
    str: string;
begin
  dt := _Data;
  if strmidx = -1 then
    st := ReadStream(_Data, Null, nil)
  else
    st := ReadStream(_Data, _data_structure[strmidx], nil);
  if st <> nil then begin
    si := 0;
    for i := 0 to high(_data_structure) do begin
      if i = strmidx then
        continue;
      tdt := ReadMTData(_Data, _data_structure[i]);
      if (Struct[si].isArray) and (Struct[si].arrlen > 1) then begin
        if Struct[si].ftype = STYPE_CHAR then begin
          str := ToString(tdt);
          if Length(str) < Struct[si].arrlen then
            SetLength(str, Struct[si].arrlen);
          st.write(str[1], Struct[si].arrlen);
        end else begin
          for j := 0 to Struct[si].arrlen-1 do begin
            WriteField(st, tdt, Struct[si]);
            if tdt.ldata <> nil then
              tdt := tdt.ldata^
            else
              dtNull(tdt);
          end;
        end;
      end else begin
        WriteField(st, tdt, Struct[si]);
      end;
      inc(si);
    end;
    _hi_CreateEvent(_Data, @onWrite, dt);
  end;
end;

end.
